extern crate actix_files;
extern crate actix_web;
extern crate serde;
extern crate serde_json;
#[macro_use]
extern crate serde_derive;
extern crate rusqlite;
extern crate time;
#[macro_use]
extern crate failure;
extern crate r2d2;
extern crate r2d2_sqlite;

use actix_files as fs;
use actix_files::NamedFile;
use actix_web::{web, App, HttpRequest, HttpResponse, HttpServer};
use failure::Error;
use r2d2_sqlite::SqliteConnectionManager;
use rusqlite::params;
use std::collections::HashMap;
use std::env;
use std::ops::Deref;
use std::sync::Mutex;
use std::time::SystemTime;

const RATE_LIMIT: i32 = 10;
const RATE_LIMIT_PER_SECOND: i32 = 60;

pub struct State {
  rate_limits: Mutex<HashMap<String, RateLimitBucket>>,
  pool: r2d2::Pool<SqliteConnectionManager>,
}

#[derive(Debug)]
pub struct RateLimitBucket {
  last_checked: SystemTime,
  allowance: f64,
}

fn main() {
  println!("Access me at {}", endpoint());

  let manager = SqliteConnectionManager::file(torrents_db_file());
  let pool = r2d2::Pool::builder().max_size(15).build(manager).unwrap();

  let shared_data = web::Data::new(State {
    rate_limits: Mutex::new(HashMap::new()),
    pool: pool,
  });

  HttpServer::new(move || {
    App::new()
      .route("/", web::get().to(index))
      .service(fs::Files::new("/static", front_end_dir()))
      .register_data(shared_data.clone())
      .route("/service/search", web::get().to(search))
  })
  .bind(endpoint())
  .unwrap()
  .run()
  .unwrap()
}

fn index() -> Result<NamedFile, actix_web::error::Error> {
  Ok(NamedFile::open(front_end_dir() + "/index.html")?)
}

fn front_end_dir() -> String {
  env::var("TORRENTS_CSV_FRONT_END_DIR").unwrap_or("../ui/dist".to_string())
}

fn torrents_db_file() -> String {
  env::var("TORRENTS_CSV_DB_FILE").unwrap_or("../../torrents.db".to_string())
}

fn endpoint() -> String {
  env::var("TORRENTS_CSV_ENDPOINT").unwrap_or("0.0.0.0:8080".to_string())
}

#[derive(Deserialize)]
struct SearchQuery {
  q: String,
  page: Option<usize>,
  size: Option<usize>,
  type_: Option<String>,
}

fn search(
  req: HttpRequest,
  data: web::Data<State>,
  query: web::Query<SearchQuery>,
) -> HttpResponse {
  let ip = req
    .connection_info()
    .remote()
    .unwrap_or("127.0.0.1:12345")
    .split(":")
    .next()
    .unwrap_or("127.0.0.1")
    .to_string();

  if query.q.is_empty() {
    return HttpResponse::BadRequest()
      .header("Access-Control-Allow-Origin", "*")
      .content_type("application/json")
      .body(format!("{{\"error\": \"{}\"}}", "Empty query".to_string()));
  }

  let conn = data.pool.get().unwrap();

  match check_rate_limit_full(data, &ip, RATE_LIMIT, RATE_LIMIT_PER_SECOND) {
    Ok(_) => HttpResponse::Ok()
      .header("Access-Control-Allow-Origin", "*")
      .content_type("application/json")
      .body(search_query(query, conn).unwrap()),
    Err(e) => HttpResponse::BadRequest()
      .header("Access-Control-Allow-Origin", "*")
      .content_type("application/json")
      .body(format!("{{\"error\": \"{}\"}}", e.to_string())),
  }
}

fn search_query(
  query: web::Query<SearchQuery>,
  conn: r2d2::PooledConnection<SqliteConnectionManager>,
) -> Result<String, serde_json::Error> {
  let page = query.page.unwrap_or(1);
  let size = query.size.unwrap_or(10);
  let type_ = query.type_.as_ref().map_or("torrent", String::deref);
  let offset = size * (page - 1);

  println!(
    "query = {}, type = {}, page = {}, size = {}",
    query.q, type_, page, size
  );

  if type_ == "file" {
    let results = torrent_file_search(conn, &query.q, size, offset).unwrap();
    serde_json::to_string(&results)
  } else {
    let results = torrent_search(conn, &query.q, size, offset).unwrap();
    serde_json::to_string(&results)
  }
}

#[derive(Debug, Serialize, Deserialize)]
struct Torrent {
  infohash: String,
  name: String,
  size_bytes: isize,
  created_unix: u32,
  seeders: u32,
  leechers: u32,
  completed: Option<u32>,
  scraped_date: u32,
}

fn torrent_search(
  conn: r2d2::PooledConnection<SqliteConnectionManager>,
  query: &str,
  size: usize,
  offset: usize,
) -> Result<Vec<Torrent>, Error> {
  let stmt_str = "select * from torrents where name like '%' || ?1 || '%' limit ?2 offset ?3";
  let mut stmt = conn.prepare(&stmt_str)?;
  let torrent_iter = stmt.query_map(
    params![
      query.replace(" ", "%"),
      size.to_string(),
      offset.to_string(),
    ],
    |row| {
      Ok(Torrent {
        infohash: row.get(0)?,
        name: row.get(1)?,
        size_bytes: row.get(2)?,
        created_unix: row.get(3)?,
        seeders: row.get(4)?,
        leechers: row.get(5)?,
        completed: row.get(6)?,
        scraped_date: row.get(7)?,
      })
    },
  )?;

  let mut torrents = Vec::new();
  for torrent in torrent_iter {
    torrents.push(torrent.unwrap());
  }
  Ok(torrents)
}

#[derive(Debug, Serialize, Deserialize)]
struct File {
  infohash: String,
  index_: u32,
  path: String,
  size_bytes: isize,
  created_unix: u32,
  seeders: u32,
  leechers: u32,
  completed: Option<u32>,
  scraped_date: u32,
}

fn torrent_file_search(
  conn: r2d2::PooledConnection<SqliteConnectionManager>,
  query: &str,
  size: usize,
  offset: usize,
) -> Result<Vec<File>, Error> {
  let stmt_str = "select * from files where path like '%' || ?1 || '%' limit ?2 offset ?3";
  let mut stmt = conn.prepare(&stmt_str).unwrap();
  let file_iter = stmt.query_map(
    params![
      query.replace(" ", "%"),
      size.to_string(),
      offset.to_string(),
    ],
    |row| {
      Ok(File {
        infohash: row.get(0)?,
        index_: row.get(1)?,
        path: row.get(2)?,
        size_bytes: row.get(3)?,
        created_unix: row.get(4)?,
        seeders: row.get(5)?,
        leechers: row.get(6)?,
        completed: row.get(7)?,
        scraped_date: row.get(8)?,
      })
    },
  )?;

  let mut files = Vec::new();
  for file in file_iter {
    files.push(file.unwrap());
  }
  Ok(files)
}

fn check_rate_limit_full(
  state: web::Data<State>,
  ip: &str,
  rate: i32,
  per: i32,
) -> Result<(), Error> {
  let mut rate_limits = state.rate_limits.lock().unwrap();

  if rate_limits.get_mut(ip).is_none() {
    rate_limits.insert(
      ip.to_string(),
      RateLimitBucket {
        last_checked: SystemTime::now(),
        allowance: -2f64,
      },
    );
  }
  if let Some(rate_limit) = rate_limits.get_mut(ip) {
    // The initial value
    if rate_limit.allowance == -2f64 {
      rate_limit.allowance = rate as f64;
    };

    let current = SystemTime::now();
    let time_passed = current
      .duration_since(rate_limit.last_checked)
      .unwrap()
      .as_secs() as f64;
    rate_limit.last_checked = current;
    rate_limit.allowance += time_passed * (rate as f64 / per as f64);
    if rate_limit.allowance > rate as f64 {
      rate_limit.allowance = rate as f64;
    }

    if rate_limit.allowance < 1.0 {
      println!(
        "Rate limited IP: {}, time_passed: {}, allowance: {}",
        &ip, time_passed, rate_limit.allowance
      );
      Err(format_err!(
        "Too many requests for IP: {}. {} per {} seconds",
        &ip,
        rate,
        per
      ))
      .unwrap()
    } else {
      rate_limit.allowance -= 1.0;
      Ok(())
    }
  } else {
    Ok(())
  }
}

#[cfg(test)]
mod tests {
  use time::PreciseTime;

  #[test]
  fn test() {
    let start = PreciseTime::now();
    let results = super::torrent_search("sherlock", 10, 0);
    assert!(results.len() > 2);
    let end = PreciseTime::now();
    println!("Query took {} seconds.", start.to(end));
  }
}
